/*
 * Original code written by Toshio Endo (endo@is.titech.ac.jp)
 */
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>

#define MAX_ITER (10000)    
#define NX (16)
#define NY (16)

long time_diff_us(struct timeval st, struct timeval et)
{
    return (et.tv_sec-st.tv_sec)*1000000+(et.tv_usec-st.tv_usec);
}

int initialize(float *region[2], int nx, int ny)
{
    int rid;
    int i, j;
    /* initialize two regions */
    for (rid = 0; rid < 2; rid++) {
        for (i = 0; i < nx; i++) {
            for (j = 0; j < ny; j++) {
                region[rid][j*nx+i] = 0.0;
            }
        }
        /* right edge */
        for (j = 0; j < ny; j++) {
            region[rid][j*nx+(nx-1)] = (float)j/(float)ny;
        }
        /* lower edge */
        for (i = 0; i < nx; i++) {
            region[rid][(ny-1)*nx+i] = (float)i/(float)nx;
        }
    }
    return 0;
}

void jacobi(float *region[2], int nx, int ny, float omega)
{
    int i, j;
    int iter;
    /* buffer id */
    int cur = 0;
    int next = 1;

    for (iter = 0; iter < MAX_ITER; iter++) {
        for (j = 1; j < ny-1; j++) {
            for (i = 1; i < nx-1; i++) {
                float curv = region[cur][j*nx+i];
                float nextv;
                /* compute new value from neighbor points */
                nextv = (region[cur][(j-1)*nx+i] +
                         region[cur][(j+1)*nx+i] +
                         region[cur][j*nx+(i-1)] +
                         region[cur][j*nx+(i+1)]) * 0.25f;
                /* overcorrection */
                region[next][j*nx+i] = curv + (nextv - curv) * omega;
            }
        }
        /* switch buffer */
        cur = 1-cur;
        next = 1-next;
    }
    return;
}

int main(int argc, char *argv[])
{
    int nx, ny;
    float omega = 0.5;
    float *region[2]; /* for double buffer */
    struct timeval st;
    struct timeval et;
    long us;

    if (argc == 3) {
        nx = atoi(argv[1]);
        ny = atoi(argv[2]);
    } else {
        nx = NX;
        ny = NY;
    }
    
    /* allocate 2D region */
    /* double buffer */
    region[0] = (float*)malloc(sizeof(float)*nx*ny);
    region[1] = (float*)malloc(sizeof(float)*nx*ny);

    initialize(region, nx, ny);

    gettimeofday(&st, NULL); /* get start time */
    jacobi(region, nx, ny, omega);
    gettimeofday(&et, NULL); /* get end time */

    us = time_diff_us(st, et);
    printf("Elapsed time of Point Jacobi of %dx%d points:"
           " %ld us\n", nx, ny, us);

    free(region[0]);
    free(region[1]);
    return 0;
}
